
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <string.h>
#include <math.h>
#include "sfftw.h"
//#include "srfftw.h"
#include "dmath.h"
#include "curvelet.h"

#ifdef __PAR__
#include <mpi.h>
#include <sfftw_mpi.h>
#endif

DCurvelet::DCurvelet(int dim)
{
  Dim = dim;
  Nscales = 0;

  Ngr=0;
  Nk[0] = Nk[1] = Nk[2] = 0;
  Lk[0] = Lk[1] = Lk[2] = 0;
  Nangles = 0;

  Angles = 0;

  Norder = 0;
  Nx = Ny = Nz = 0;
  
  CoordWinCenter = 0;
  OversampleFac = 0;

  US_OverSample = US_OVERSAMPLE_DEFAULT;
  US_Epsilon = US_EPSILON_DEFAULT;

  WindowData = 0;
}

DCurvelet::~DCurvelet()
{

  if(Angles) DMath::del(Angles);
  if(OversampleFac) DMath::del(OversampleFac);

  if(WindowData) delete WindowData;
}


int DCurvelet::SetUp(int nx, int ny, int nz, int over_sample, double eps, int flag, int flag1)
{
  Nx = nx;
  Ny = ny;
  Nz = nz;
  Nm = max(Nx,max(Ny,Nz));

  //**************************************************//

  US_OverSample = over_sample;
  US_Epsilon = eps;

  double logeps = -log(US_Epsilon);
  US_lambda = PI*PI/(2.0*logeps);
  US_d_eps = (int)(sqrt(logeps/ US_lambda)); 


  WindowData = new DBox();
  WindowData->SetUp(Nm, eps, flag, flag1);


  Nscales = WindowData->Kend;
 
  Lk[0] = WindowData->L1; // window dimensions
  Lk[1] = WindowData->L2;
  Lk[2] = WindowData->L2;

  Nangles = WindowData->N_nu; 
  CoordWinCenter = WindowData->Kc; 

  NumAnglesTot = WindowData->NboxTot;

  Ngr = WindowData->Ngr;
  Nk[0] = WindowData->Nk[0];
  Nk[1] = WindowData->Nk[1];
  Nk[2] = WindowData->Nk[2];

  return 0;
}



int DCurvelet::ReadDataBlock(fftw_complex ***block, int *Pbeg, int *Pend) 
{
  int m = US_OverSample;
  int Nxm = m*Nx, Nym = m*Ny,  Nzm = m*Nz;
  int i, j; //ind, iter, i, j, kk, lx, ly, lz;
  FILE *fid;

  int Nx1=Pbeg[0]+Nxm/2, Ny1=Pbeg[1]+Nym/2, Nz1=Pbeg[2]+Nzm/2, Nx2=Pend[0]+Nxm/2, Ny2=Pend[1]+Nym/2, Nz2=Pend[2]+Nzm/2;
  int dNx=Nx2-Nx1, dNy=Ny2-Ny1, dNz=Nz2-Nz1;  
cout << ":::::::::::::::::::::Pbeg: " << Pbeg[0] << ", "<< Pbeg[1] << ", " << Pbeg[2] << "\n";  cout.flush();  
cout << ":::::::::::::::::::::Pend: " << Pend[0] << ", "<< Pend[1] << ", " << Pend[2] << "\n";  cout.flush();  
cout << ":::::::::::::::::::::Nx1...: " << Nx1 << ", "<< Ny1 << ", " << Nz1 << "\n";  cout.flush();  
cout << ":::::::::::::::::::::Nx2...: " << Nx2 << ", "<< Ny2 << ", " << Nz2 << "\n";  cout.flush();  
cout << ":::::::::::::::::::::dNx...: " << dNx << ", "<< dNy << ", " << dNz << "\n";  cout.flush();  

 long strides[3]; 
  strides[0]=Nz1+Ny1*Nzm+Nx1*Nym*Nzm; strides[1]=(Nym-dNy)*Nzm; strides[2]=Nzm-dNz; 
cout << "strides: " << strides[0] << ", "<< strides[1] << ", " << strides[2] << "\n";  cout.flush();   
  fid = fopen(INP_FFT_FILE,"rb");
  fseek (fid , strides[0]*sizeof(fftw_complex), SEEK_SET);
  for(i=0;i<dNx;i++) {  
        for(j=0;j<dNy;j++)  {
             fread(&block[i][j][0], sizeof(fftw_complex), dNz, fid);
             fseek (fid , strides[2]*sizeof(fftw_complex), SEEK_CUR);
         } // for(j=0...
         fseek (fid , strides[1]*sizeof(fftw_complex), SEEK_CUR);
  } // for(i=0...
  fclose (fid);
cout << ":::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n";  cout.flush();  

  return 0;
}

int DCurvelet::WriteDataBlock(fftw_complex ***block, int *Pbeg, int *Pend, int flag_add) 
{          // flag_add=1 means that we read a block from file add it with our one and then write it back to file
            // flag_add=0 means that we just overwrite data.
  int m = US_OverSample;
  int Nxm = m*Nx, Nym = m*Ny,  Nzm = m*Nz;
  int i, j, kk; //ind, iter, i, j, kk, lx, ly, lz;
  FILE *fid;

  int Nx1=Pbeg[0]+Nxm/2, Ny1=Pbeg[1]+Nym/2, Nz1=Pbeg[2]+Nzm/2, Nx2=Pend[0]+Nxm/2, Ny2=Pend[1]+Nym/2, Nz2=Pend[2]+Nzm/2;
  int dNx=Nx2-Nx1, dNy=Ny2-Ny1, dNz=Nz2-Nz1;  

cout << "Write:::::::::::::::::::::Pbeg: " << Pbeg[0] << ", "<< Pbeg[1] << ", " << Pbeg[2] << "\n";  cout.flush();  
cout << "Write:::::::::::::::::::::Nx1...: " << Nx1 << ", "<< Ny1 << ", " << Nz1 << "\n";  cout.flush();  
cout << "Write:::::::::::::::::::::Nx2...: " << Nx2 << ", "<< Ny2 << ", " << Nz2 << "\n";  cout.flush();  
cout << "Write:::::::::::::::::::::dNx...: " << dNx << ", "<< dNy << ", " << dNz << "\n";  cout.flush();  

  fftw_complex *buf; buf=(fftw_complex*)malloc(sizeof(fftw_complex)*dNz);
  int strides[3];  strides[0]=Nz1+Ny1*Nzm+Nx1*Nym*Nzm; strides[1]=(Nym-dNy)*Nzm; strides[2]=Nzm-dNz; 
    
if(flag_add==1) {
  fid = fopen(OUTP_FFT_FILE,"rb+");
  fseek (fid , strides[0]*sizeof(fftw_complex), SEEK_SET);
  for(i=0;i<dNx;i++) {  
        for(j=0;j<dNy;j++)  {
             fread(&buf[0], sizeof(fftw_complex), dNz, fid);
             for(kk=0;kk<dNz;kk++)  { buf[kk].re+=block[i][j][kk].re; buf[kk].im+=block[i][j][kk].im; }
             fseek (fid , -dNz*sizeof(fftw_complex), SEEK_CUR);
             fwrite(&buf[0], sizeof(fftw_complex), dNz, fid);
             fseek (fid , strides[2]*sizeof(fftw_complex), SEEK_CUR);
         } // for(j=0...
         fseek (fid , strides[1]*sizeof(fftw_complex), SEEK_CUR);
  } // for(i=0...
  fclose (fid);
} // if(flag_add=...
else {
  fid = fopen(OUTP_FFT_FILE,"rb+");
  fseek (fid , strides[0]*sizeof(fftw_complex), SEEK_SET);
  for(i=0;i<dNx;i++) {  
        for(j=0;j<dNy;j++)  {
             fwrite(&block[i][j][0], sizeof(fftw_complex), dNz, fid);
             fseek (fid , strides[2]*sizeof(fftw_complex), SEEK_CUR);
         } // for(j=0...
         fseek (fid , strides[1]*sizeof(fftw_complex), SEEK_CUR);
  } // for(i=0...
  fclose (fid);
} // else...

  free(buf);
  return 0;
}



#ifdef __PAR__

int DCurvelet::parWriteCcoeff(char *FileName, char *FileName1, char *FileName2, vector<fftw_complex> *vectv, vector<int> *vect1, int *cc_angles, int Nangles, MPI_Comm comm)
{
   int my_rank, ncpus;    MPI_Comm_rank(comm, &my_rank);  MPI_Comm_size(comm, &ncpus);
   MPI_Status status;
   FILE *fid;
   int ccc, ii, nummax; 
   int *wbind,  *wbnum, *cc_num, *cc_info; //*dangls; 
   fftw_complex *wbval;
   cc_num=(int *)malloc(ncpus*sizeof(int));  // number of non-zero coefficients from each cpu
   cc_info= (int *)malloc(Nangles*sizeof(int));

    MPI_Reduce(&cc_angles[0], &cc_info[0], Nangles, MPI_INT, MPI_SUM, (ncpus-1), MPI_COMM_WORLD);          
    ccc=(int)vect1->size();
    MPI_Gather(&ccc, 1, MPI_INT, &cc_num[0], 1, MPI_INT, (ncpus-1), MPI_COMM_WORLD);


   if( my_rank != (ncpus-1))    if( (int)vect1->size() != 0) { 
               MPI_Send(&vect1->at(0), (int)vect1->size(), MPI_INT, (ncpus-1), 98, MPI_COMM_WORLD); 
               MPI_Send(&vectv->at(0), 2*(int)vect1->size(), MPI_FLOAT, (ncpus-1), 99, MPI_COMM_WORLD); 
    }   

      if( my_rank == (ncpus-1)) { 
            nummax=0; for(ii=0; ii< (ncpus-1); ii++)   if(cc_num[ii]>nummax) nummax=cc_num[ii]; 
            wbind=(int *)malloc(nummax*sizeof(int));  wbval=(fftw_complex *)malloc(nummax*sizeof(fftw_complex));
            fid = fopen(FileName,"wb"); fclose (fid); fid = fopen(FileName1,"wb"); fclose (fid); 
 
           for(ii=0; ii< (ncpus-1); ii++)   if( cc_num[ii] != 0) { 
                  MPI_Recv(&wbind[0], cc_num[ii], MPI_INT, ii, 98, MPI_COMM_WORLD, &status);
                  fid = fopen(FileName,"ab"); fwrite(&wbind[0], sizeof(int), cc_num[ii], fid);  fclose (fid); 
                  MPI_Recv(&wbval[0], 2*cc_num[ii], MPI_FLOAT, ii, 99, MPI_COMM_WORLD, &status);
                  fid = fopen(FileName1,"ab"); fwrite(&wbval[0], sizeof(fftw_complex), cc_num[ii], fid);  fclose (fid);  
            }  // for(ii=0; ii< (ncpus-1).....
            if( (int)vect1->size() != 0) { 
                fid = fopen(FileName,"ab"); fwrite(&vect1->at(0), sizeof(int), (int)vect1->size(), fid);  fclose (fid);
                fid = fopen(FileName1,"ab"); fwrite(&vectv->at(0), sizeof(fftw_complex), (int)vect1->size(), fid);  fclose (fid);
            }
            fid = fopen(FileName2,"wb"); fwrite(&cc_info[0], sizeof(int), Nangles, fid);  fclose (fid); 
            free(wbind);  free(wbval);        
     } //  if( my_rank == ......


  free(cc_num);  free(cc_info);
  return 0;
}


int DCurvelet::parReadCcoeff(char *FileName, char *FileName1, char *FileName2, vector<fftw_complex> *vectv, vector<int> *vect1, int *cc_angles, int Nangles, MPI_Comm comm)
{
   int my_rank, ncpus;    MPI_Comm_rank(comm, &my_rank);  MPI_Comm_size(comm, &ncpus);
   MPI_Status status;
   FILE *fid;
   int ii, jj, nummax, shift; 
   int *wbind,  *cc_num; 
   fftw_complex *wbval;
   fftw_complex  zer;   zer.im=zer.re=0.0; 
   cc_num=(int *)malloc(ncpus*sizeof(int));  // number of non-zero coefficients from each cpu

    int res, num, angl1, dangl1;
    res=Nangles%ncpus; num=Nangles/ncpus;
      
    if( my_rank == (ncpus-1))   {
           fid = fopen(FileName2,"rb"); fread(&cc_angles[0], sizeof(int), Nangles, fid);  fclose (fid); 
    }
     MPI_Bcast(&cc_angles[0],Nangles, MPI_INT, (ncpus-1), MPI_COMM_WORLD);
     for(ii=0, angl1=0; ii< ncpus; ii++) {
          if(ii<res)  dangl1=num+1;  else dangl1=num;
          for(jj=angl1, cc_num[ii]=0; jj< (angl1+dangl1); jj++) cc_num[ii]+=cc_angles[jj];  angl1+=dangl1;
     } // for(ii=...

     vect1->assign(cc_num[my_rank], 0);   vectv->assign(cc_num[my_rank], zer);

     if( my_rank == (ncpus-1)) { 
          nummax=0; for(ii=0; ii< (ncpus-1); ii++) if(cc_num[ii]>nummax) nummax=cc_num[ii]; 
          wbind=(int *)malloc(nummax*sizeof(int));  wbval=(fftw_complex *)malloc(nummax*sizeof(fftw_complex));
      
          for(ii=0, shift=0; ii< (ncpus-1); ii++)    if( cc_num[ii] != 0) { 
                fid = fopen(FileName,"rb");  fseek (fid, shift*sizeof(int), SEEK_SET); fread(&wbind[0], sizeof(int), cc_num[ii], fid);  fclose (fid);       
                MPI_Send(&wbind[0], cc_num[ii], MPI_INT, ii, 98, MPI_COMM_WORLD); 
                fid = fopen(FileName1,"rb"); fseek (fid, shift*sizeof(fftw_complex), SEEK_SET); fread(&wbval[0], sizeof(fftw_complex), cc_num[ii], fid);  fclose (fid); 
                MPI_Send(&wbval[0], 2*cc_num[ii], MPI_FLOAT, ii, 99, MPI_COMM_WORLD); 
                shift+=cc_num[ii];
          } // for(ii=0; ii< (ncpus-1).....         
          if( cc_num[my_rank] != 0) { 
               fid = fopen(FileName,"rb");  fseek (fid, shift*sizeof(int), SEEK_SET); fread(&vect1->at(0), sizeof(int), cc_num[my_rank], fid);  fclose (fid);       
               fid = fopen(FileName1,"rb"); fseek (fid, shift*sizeof(fftw_complex), SEEK_SET); fread(&vectv->at(0), sizeof(fftw_complex), cc_num[my_rank], fid);  fclose (fid); 
          }
          free(wbind); free(wbval); 
     } //  if( my_rank == (ncpus-1) .........

     if( my_rank != (ncpus-1))   if( cc_num[my_rank] != 0) { 
            MPI_Recv(&vect1->at(0), cc_num[my_rank], MPI_INT, (ncpus-1), 98, MPI_COMM_WORLD, &status);
            MPI_Recv(&vectv->at(0), 2*cc_num[my_rank], MPI_FLOAT, (ncpus-1), 99, MPI_COMM_WORLD, &status); 
     } 

    free(cc_num);  
    return 0;
}


#endif  // __PAR__




double DCurvelet::InputData(char *input_file, fftw_complex ***data,  int nx, int ny, int nz)
{
  fstream DataIn(input_file, ios::in | ios::binary);
  if(!DataIn) {
    cout << "Error: Input file " << input_file << " open failed!\n";
    return -1;
  }

  float *dData = new float[nx*ny*nz];
  double tmp, max; 

  DataIn.read((char*)dData, sizeof(float)*nx*ny*nz);

  double *padx = new double[nx]; double *pady = new double[ny];  double *padz = new double[nz];
  int i,j,k;

  for(i=0;i<nx;i++) padx[i]=1.0;  for(i=0;i<ny;i++) pady[i]=1.0;  for(i=0;i<nz;i++) padz[i]=1.0;
//  for(i=0;i<8;i++) { padx[i]=0.0+i*1/8.0; padx[nx-1-i]=0.0+i*1/8.0; pady[i]=0.0+i*1/8.0; pady[ny-1-i]=0.0+i*1/8.0; padz[i]=0.0+i*1/8.0; padz[nz-1-i]=0.0+i*1/8.0; }

  for(i=0;i<nx;i++)
    for(j=0;j<ny;j++)
      for(k=0;k<nz;k++) {
        tmp=(double)(dData[(i*ny+j)*nz+k])*pady[i]*padx[j]*padz[k];
        data[i][j][k].re = tmp;   data[i][j][k].im = 0.0;
        if(fabs(tmp)>max) max=fabs(tmp);
      }

   cout << "max value " << max << " \n";
  tmp=1.0/max;
  for(i=0;i<nx;i++)  for(j=0;j<ny;j++)  for(k=0;k<nz;k++) data[i][j][k].re*=tmp;

  // clean up
  delete []dData;
  delete padx; delete pady; delete padz; 
 
  return max;
}



