#ifdef __PAR__
#include <mpi.h>
#endif

#include "sfftw.h"
#include <vector>
using std::vector;


#ifndef CURVELET_H
#define CURVELET_H

#ifndef max
#define max(a,b) ((a)>(b)?(a):(b)) // max of a, b
#endif

#ifndef min
#define min(a,b) ((a)<(b)?(a):(b)) // min of a, b
#endif


#ifndef PI
#define PI 3.1415926535897932
#endif

    // Supposed to be modified 
#define IMAGE_SIZE   128 // image size 
#define INTERNAL_PENALTY 0.001  // penalty used to threshold coefficients in 'exact' transform 
// #define INTERNAL_PENALTY 0.001  // penalty used to threshold coefficients in 'exact' transform 
#define MEMORY_PER_NODE  2000  // maximum memory avilable per CPU !!!!! 
#define US_EPSILON_DEFAULT    1.0e-5  

    // Supposed to be modified 
#define US_OVERSAMPLE_DEFAULT  2.0     
#define MAX_REDUCE_BUF_SIZE  100 // maximum memory size for buffer used by Reduce() during inverse transform;
#define INP_FFT_FILE "fft_spectr.in" // weighted FFT spectrum (input to smearing in curvelet decomposition)
#define OUTP_FFT_FILE "fft_spectr.out" // weighted FFT spectrum (output from smearing in curvelet recomposition)



class DBox 
{
 public:
  DBox();
  ~DBox();

// BLOCK 0 - input parameters

 public:
  int DIM;       // dimensionality of the transform (3D)
  double a0;     // radius of the coarsest scale (k=0)
  double nn0;    // number of angles for first scale (k=1)

// BLOCK 1  - parameters defining box geometry and orientations

  int Kend;      // maximum number of scales

 public:
  double *K1;    // box pseudo center (coeff*2^scale)  
  int *N_nu_c;   // ...(internal)... number of angles in one direction per sector of cubed sphere
  int *N_nu_sym; // number of different window functions (after accounting for all symmetries)
  long *N_nu;    // total number of angles for each scale
  double *L1;    // box length
  double *L2;    // box width
  double *K1low; // box lower boundary (upper boundary is 2*K1)
                 // lower boundary is different from K1/2 due to use of polar coordinates
  double *Kc;    // box actual center (between K1low and 2*K1)


                  // box orientations (stored in one vector for all scales)
 public:          // separate scales can be found using N_nu array
  double **boxw1; // radial direction
  double **boxw2; // direction of the first short side
  double **boxw3; 

// BLOCK 2 - grid and window function on boxes

 public:
  double **OverSampleFac; // array of oversampling factors ... doubt that they are used in computations <produced by "setup_Nk()">
  int **Nk;              // array containing grid size (nx,ny,nz) for boxes at different scales <produced by "setup_Nk()">
  long *Ngr;              // array containing box size nx*ny*nz for boxes at different scales <produced by "setup_Nk()">
  float ***WindowFunc;   // !! huge array with window function values for every box  <produced by "setup_window_func()"> 
  float **WindowFuncScale;   // !! huge array with window function values for every box  <produced by "setup_window_func()"> 
  double ***Grid[3];      // !! huge array with global coordinates of grid points for every box <produced by "setup_grids()">
  int **Pqs_min;  // array containing minimum coord in (x,y,z) directions for region containing boxes of first quater sector <produced by "setup_window_func()">
  int **Pqs_max;  // array containing maximum coord in (x,y,z) directions for region containing boxes of first quater sector <produced by "setup_window_func()">
 
 public:
  long NboxTot;           // total number of boxes at all scales (size of N_nu) produced by "setup_Nk()"
  long *NgridsScale;    // number of grid points (curv. coeff.) per scale
  long NgridsTot;         // total number of grid points = NUMBER OF CURVELET COEFFICIENTS
  long *NgridsWinScale;   // number of window function points per scale
  long NgridsWinTot;        // total number of window function points
  long MaxNgridsPerBox;   // the size of the biggest grid (size of the grid for box at finest scale I guess) 
 
// BLOCK 3 - flags showing existence of window function array and grid array
 public:
  int isWinFuncSetup; 
  int isGridSetup; 

// BLOCK 4 - parallel: flags for memory constraints

  int flagp2; // check if it is possible to work with a sector locally; 
  int *flagp3; // check if it is possible to work with one box locally; 
  int KK0, KK1, KK2; // scale for which we can can fit in memory: whole scale, quadrant, quater-sector. 
  double Lamd; 
  int half_q; 

// MAIN FUNCTIONS - for full curvelet transform 

// Main function to initialize object of class DBox
// (flag gives you an option not to precompute window function arrays) 
 public:  
  int SetUp(int N_image, double eps=US_EPSILON_DEFAULT, int flag=1, int flag1=0); 

          // allocating memory for a number of boxes: form k=0 up to k=nscales;  
 public:
    float*** allocD(int nscales, int *nangles, int **ngrids);

// Internal functions mainly used to execute SetUp()

 private:
  int Cube_sphere_1(int NN, int dim, double *w1, double *w2, double *w3, double *r_box);

public: 
 long find_opposite(int scale, long snu);  
 long find_closest(int scale, double *vect); 

 public:
 void apol(double* t,double *y,int N);
 void meyer(double *t, double *y, int N,int type);

 void chihat_ang(float *x1, float *x2, float *x3, double *chihat, int *N, double *L, double K1);
 void chihat_rad(float *x1, float *x2, float *x3, double *chihat, int *N, double K1, int type);

 void chihat_f(float *x1, float *x2, float *x3,  double *boxw1, double *boxw2, double *boxw3, 
  	     float *w, int *N, double *L, double Scale_rad, int Np, int kk, int snu); // saving window function as float (memory saving mode)

 void chihat_sm(float *x1, float *x2, float *x3,  double *boxw1, double *boxw2, double *boxw3, 
	      double *w, int *N, double *L, double K1, int Np, int scale, int snu);  
 void chihat_sm_f(float *x1, float *x2, float *x3,  double *boxw1, double *boxw2, double *boxw3, 
	      float *w, int *N, double *L, double K1, int Np, int scale, int snu);  


  void setup_Nk(double eps=US_EPSILON_DEFAULT);
  void get_os(int scale, double *L, double eps, int *N, double *over_sample);

  void setup_window_func();
  int GetWindowFunctScale(int k1); // read window function from file for one scale


// MAIN FUNCTIONS for full transform

 public:
// TEST FUNCTIONS - working with boxes individually 
//                  for testing and research regime

 public:   //to calculate window function for individual box
  int CalcWindowFunc_f(int k, int ix_angle, float *window_func); // saving window function as float (memory saving mode)
 
  int CalcGrid(int k, int ix_angle, float *x1, float *x2, float *x3);
  int CalcGrid(int k, float *yx1, float *yx2, float *yx3);
};




class DCurvelet
{
 public:
  DCurvelet(int dim=3);
  ~DCurvelet();
   
 public:
  int Dim; 
  int Nscales; 
   
  int *Nk[3]; 
  long *Ngr;              // array containing box size nx*ny*nz for boxes at different scales <produced by "setup_Nk()">

  double *Lk[3]; 

  long *Nangles; 
  double **Angles; 

  long NumAnglesTot; 
  int Norder; 

  double *CoordWinCenter; 
  double *OversampleFac; 

  int US_OverSample; 
  double US_Epsilon; 
  double US_lambda; 
  int US_d_eps; 

                                 // arrays for writing curvelet coefficients
 public:
  DBox *WindowData; 

 public:
  int Nx, Ny, Nz; 
  int Nm; 

 public:

   int SetUp(int nx, int ny, int nz, int over_sample=US_OVERSAMPLE_DEFAULT, double eps=US_EPSILON_DEFAULT, int flag=1, int flag1=0); 

   int GetPerfectReconst(char *input_file, int Nx, int Ny, int Nz);

   int TransformScaleQuadPar(vector<fftw_complex> *vectv, vector<int> *vect1,  int *cc_angles, fftw_complex ***fft_image, float **WindFuncsAll, int *Nbeg, int *Nend, int k, int ang1, int dang1, float penalty);
   int InvTransformScaleQuadPar(vector<fftw_complex> *vectv, vector<int> *vect1,  int *cc_angles, fftw_complex ***fft_image, float **WindFuncsAll, int *Nbeg, int *Nend, int k, int ang1, int dang1);
 
  int seqIFFT(fftw_complex ***func, fftw_complex ***fftCoeff);
  int seqFFT(fftw_complex ***fftData, fftw_complex ***func);

#ifdef __PAR__
  int parFFT(char *input_file, char *max_v_file, MPI_Comm comm);
  int parIFFT(char *output_file, char *max_v_file, MPI_Comm comm); 
#endif

  int ReadDataBlock(fftw_complex ***block, int *Pbeg, int *Pend);
  int WriteDataBlock(fftw_complex ***block, int *Pbeg, int *Pend, int flag_add); 

#ifdef __PAR__
   int parBcastRedMem(fftw_complex ***block, int dNx, int dNy, int dNz, MPI_Comm comm);
   int parReduceRedMem(fftw_complex ***imag_rec, int dNx, int dNy, int dNz, MPI_Comm comm); 

   int parWriteCcoeff(char *FileName, char *FileName1, char *FileName2, vector<fftw_complex> *vectv, vector<int> *vect1, int *cc_angles, int Nangles, MPI_Comm comm);
   int parReadCcoeff(char *FileName, char *FileName1, char *FileName2, vector<fftw_complex> *vectv, vector<int> *vect1, int *cc_angles, int Nangles, MPI_Comm comm);
#endif

 private:

  int E2UBox(fftw_complex *fftBox, fftw_complex ***fftData, int *Nst, int *Nend, int k, int ix_angle);
//  int E2UBox(fftw_complex *fftBox, fftw_complex ***fftData, int k, int ix_angle);
  int U2EBox(fftw_complex ***fftData, fftw_complex *fftBox, int *Nst, int *Nend, int k, int ix_angle, int flag_addimage);
//  int U2EBox(fftw_complex ***fftData, fftw_complex *fftBox, int k, int ix_angle, int flag_addimage);

 private:
  int fftshift3(fftw_complex *f, int* lengths, int l);
  int fftshift2(fftw_complex *f, int* lengths, int l);
  void get_gauss_bell(double b, int q, double *coeff);
  void get_gauss_bell(double xi, int q1, int q2, double *gxl);
  void get_normal_coeff(int m, int N, double *gjx);
  void get_normal_coeff(int m, int N, int Nst, int Nend, double *gjx);


 private:


  double InputData(char *input_file, fftw_complex ***data, int nx, int ny, int nz);

};

#endif 




 
