#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <math.h>
#include <mpi.h>
#include "dmath.h"
#include "curvelet.h"
#include <vector>

int main(int argc, char **argv)
{
   int my_rank, ncpus;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &ncpus);
  MPI_Status status;

  int Nx, Ny, Nz=IMAGE_SIZE;
  Nx = Ny = Nz;

//  int K_Index, AngleIndex;
  char *InputFile = argv[1];

  int it, ii, jj, kk, ll, mm, mq;
  FILE *fid;
  char FileName[1024], FileName1[1024], FileName2[1024];
  float ttre;
  double t1, max_v;

 t1=MPI_Wtime();
// curvelet transform object
  DCurvelet *curv=0;
  curv = new DCurvelet();
  
  curv->SetUp(Nx, Ny, Nz, 2, US_EPSILON_DEFAULT, 0,0); // we do not allocate window function

  cout << "Setup: done... " << my_rank << "\n"; cout.flush();


// ==============  PAR - Global FFT ===================


if(1>0) {

   MPI_Group  all_cpus, fft_cpus; 
   MPI_Comm FFT_Comm; 
   int fftncpus, tmp, color, n_rank;
   char mvFile1[1024]; sprintf(mvFile1, "./CCoeff/%s_max_v",InputFile);

   tmp=(int)(log((double)ncpus)/log(2.0)); 
   fftncpus=(int)pow(2,tmp);   //cout << "FFT ncpus:" << fftncpus << "\n"; cout.flush(); 
   if(my_rank<fftncpus)  color=1; else color=0;
   MPI_Comm_split( MPI_COMM_WORLD, color, my_rank, &FFT_Comm);
   MPI_Comm_rank(MPI_COMM_WORLD, &n_rank);   

   if(my_rank<fftncpus)     curv->parFFT(InputFile, mvFile1, FFT_Comm);
}

MPI_Barrier(MPI_COMM_WORLD); // this barrier is neccessary (do not remove !)
cout << "----> Global FFT: ...done\n"; cout.flush();
MPI_Barrier(MPI_COMM_WORLD);


// ==============  PAR - Smearing out ===================

if(1>0) {  // compute all scales in one run
  vector<int> vect1; vector<fftw_complex> vectv;
  int *cc_angles; 

  int ccc, angl1, dangl1, NNangles, Nangles, res, num;
  int PPbeg[3], PPend[3];
  int dNx, dNy, dNz; 
  double vx, vy, vz; int Ps, Pe; 
  int *wbind, *cc_info; 

  fftw_complex *wbval;
                          
      // --- analyze configuration for each scale 
   int KK0=curv->WindowData->KK0, KK1=curv->WindowData->KK1, KK2=curv->WindowData->KK2;

   PPbeg[0]=PPbeg[1]=PPbeg[2]=-curv->WindowData->Pqs_max[KK0][3]; PPend[0]=PPend[1]=PPend[2]=curv->WindowData->Pqs_max[KK0][3];
   dNx=PPend[0]-PPbeg[0]; dNy=PPend[1]-PPbeg[1]; dNz=PPend[2]-PPbeg[2];  
   fftw_complex ***fft_image = DMath::newC(dNx, dNy, dNz);

   if( my_rank == (ncpus-1))   curv->ReadDataBlock(fft_image, PPbeg, PPend);
 
         // broadcast with small memory buffer
   curv->parBcastRedMem(fft_image, dNx, dNy, dNz, MPI_COMM_WORLD);


   if( my_rank == (ncpus-1)) {     // curv. coeff. for scale k=0
         curv->WindowData->WindowFuncScale = DMath::newF(curv->WindowData->N_nu_sym[0], curv->WindowData->Ngr[0]); 
         curv->WindowData->GetWindowFunctScale(0);
  
         cc_angles=(int *)malloc(1*sizeof(int));  memset(&cc_angles[0], '\0', sizeof(int)*1);

         curv->TransformScaleQuadPar(&vectv, &vect1,  &cc_angles[0], fft_image, curv->WindowData->WindowFuncScale, PPbeg, PPend, 0, 0, 1, INTERNAL_PENALTY);        

          sprintf(FileName, "./CCoeff/%s_cc_ind-k%d", InputFile, 0); 
          fid = fopen(FileName,"wb");  fwrite(&vect1[0], sizeof(int), (int)vect1.size(), fid); fclose (fid); 
          sprintf(FileName, "./CCoeff/%s_cc_val-k%d", InputFile, 0); 
          fid = fopen(FileName,"wb"); fwrite(&vectv[0], sizeof(fftw_complex), (int)vect1.size(), fid);  fclose (fid); 
          sprintf(FileName, "./CCoeff/%s_cc_info-k%d", InputFile, 0); 
          ccc=(int)vect1.size();
          fid = fopen(FileName,"wb");  fwrite(&ccc, sizeof(int), 1, fid);  fclose (fid);    
 
          vect1.clear(); vectv.clear();   free(cc_angles);
   }  // if( my_rank =...


// MPI_Barrier(MPI_COMM_WORLD);
// cout << "====> Broadcast: ...done\n"; cout.flush();
// MPI_Barrier(MPI_COMM_WORLD);


   for(mm=1; mm<=KK0; mm++) {
         Nangles=curv->WindowData->N_nu[mm]/2;   
         res=Nangles%ncpus; num=Nangles/ncpus;
         if(my_rank<res)  dangl1=num+1;  else dangl1=num;
         angl1=0;  for(ii=0; ii< my_rank; ii++)  if(ii<res) angl1+=num+1; else angl1+=num;

         cc_angles=(int *)malloc(Nangles*sizeof(int));  memset(&cc_angles[0], '\0', sizeof(int)*Nangles);

                              // allocate, read and distribute window functions
         curv->WindowData->WindowFuncScale = DMath::newF(curv->WindowData->N_nu_sym[mm], curv->WindowData->Ngr[mm]); 
         if( my_rank == (ncpus-1))  curv->WindowData->GetWindowFunctScale(mm);
         MPI_Bcast(&curv->WindowData->WindowFuncScale[0][0], curv->WindowData->NgridsWinScale[mm], MPI_FLOAT, (ncpus-1), MPI_COMM_WORLD);

         curv->TransformScaleQuadPar(&vectv, &vect1, &cc_angles[angl1], fft_image, curv->WindowData->WindowFuncScale, PPbeg, PPend, mm, angl1, dangl1, INTERNAL_PENALTY);        

         sprintf(FileName, "./CCoeff/%s_cc_ind-k%d", InputFile, mm);  sprintf(FileName1, "./CCoeff/%s_cc_val-k%d", InputFile, mm);  sprintf(FileName2, "./CCoeff/%s_cc_info-k%d", InputFile, mm);  

         curv->parWriteCcoeff(FileName, FileName1, FileName2, &vectv, &vect1, cc_angles, Nangles, MPI_COMM_WORLD);

         DMath::del(curv->WindowData->WindowFuncScale);
         vect1.clear(); vectv.clear();   free(cc_angles); free(cc_info); 
   }  //  for(mm=1; mm...


  DMath::del(fft_image);
}


MPI_Barrier(MPI_COMM_WORLD);
cout << "FORWARD Smearing: ... done.\n"; cout.flush();
MPI_Barrier(MPI_COMM_WORLD);

MPI_Barrier(MPI_COMM_WORLD);
cout << "=====================FORWARD TRANSFORM DONE=============\n"; cout.flush();
MPI_Barrier(MPI_COMM_WORLD);



// ==============  Par - Smearing in (Inverse) ===================

if(1>0) {  // compute all scales in one run
   vector<int> ivect1; vector<fftw_complex> ivectv;
   int *icc_angles; 

           // create an empty file for spectrum
   fftw_complex *tmp = DMath::newC(2*Ny*2*Nz);  memset(&tmp[0], '\0', sizeof(fftw_complex)*(2*Ny*2*Nz));
   fid = fopen("fft_spectr.out","wb");  for(ii=0;ii<(2*Nx);ii++) fwrite(&tmp[0], sizeof(fftw_complex), (2*Ny*2*Nz), fid); fclose (fid); 
   DMath::del(tmp);

  int ccc, angl1, dangl1, Nangles, NNangles, res, num, ccnonzero;
  int shift, qshift=0;
  int PPbeg[3], PPend[3];
  double ivx, ivy, ivz; int iPs, iPe; 
  int dNx, dNy, dNz; 
  int *wbind, *cc_info; 
  fftw_complex *rbval;
  fftw_complex  zer; zer.im=zer.re=0.0; 

  int KK0=curv->WindowData->KK0;

   PPbeg[0]=PPbeg[1]=PPbeg[2]=-curv->WindowData->Pqs_max[KK0][3]; PPend[0]=PPend[1]=PPend[2]=curv->WindowData->Pqs_max[KK0][3];
   dNx=PPend[0]-PPbeg[0]; dNy=PPend[1]-PPbeg[1]; dNz=PPend[2]-PPbeg[2];  
   fftw_complex ***imag_rec = DMath::newC(dNx, dNy, dNz);
        memset(&imag_rec[0][0][0], '\0', sizeof(fftw_complex)*(dNx*dNy*dNz));


   if( my_rank == (ncpus-1)) {   // scale k = 0 

       curv->WindowData->WindowFuncScale = DMath::newF(curv->WindowData->N_nu_sym[0], curv->WindowData->Ngr[0]); 
       curv->WindowData->GetWindowFunctScale(0);

       sprintf(FileName, "./CCoeff/%s_cc_info-k%d", InputFile, 0); 
       fid = fopen(FileName,"rb");  fread(&ccc, sizeof(int), 1, fid);  fclose (fid); 
       ivect1.assign(ccc, 0);    ivectv.assign(ccc, zer);

       sprintf(FileName, "./CCoeff/%s_cc_ind-k%d", InputFile, 0);   sprintf(FileName1, "./CCoeff/%s_cc_val-k%d", InputFile, 0);  
       fid = fopen(FileName,"rb");  fread(&ivect1[0], sizeof(int), ccc, fid);  fclose (fid);       
       fid = fopen(FileName1,"rb"); fread(&ivectv[0], sizeof(fftw_complex), ccc, fid);  fclose (fid); 

       curv->InvTransformScaleQuadPar(&ivectv, &ivect1, &ccc, imag_rec, curv->WindowData->WindowFuncScale, PPbeg, PPend, 0, 0, 1);        
       for(ii=0; ii< dNx; ii++)    for(jj=0; jj<dNy; jj++)    for(kk=0; kk<dNz; kk++) {
                 ll=kk+dNz*(jj+ii*dNy);  imag_rec[ii][jj][kk].re=imag_rec[ii][jj][kk].re/2.0; imag_rec[ii][jj][kk].im=imag_rec[ii][jj][kk].im/2.0;  }

        DMath::del(curv->WindowData->WindowFuncScale);
   } //  if( my_rank == (ncpus-1) .........


  for(mm=1; mm<=KK0; mm++) {
        Nangles=curv->WindowData->N_nu[mm]/2;   
        res=Nangles%ncpus; num=Nangles/ncpus;
        if(my_rank<res)  dangl1=num+1;  else dangl1=num;
        for(ii=0, angl1=0; ii< my_rank; ii++)  if(ii<res) angl1+=num+1; else angl1+=num;
        icc_angles=(int *)malloc(Nangles*sizeof(int)); 

        curv->WindowData->WindowFuncScale = DMath::newF(curv->WindowData->N_nu_sym[mm], curv->WindowData->Ngr[mm]); 
        if( my_rank == (ncpus-1))  curv->WindowData->GetWindowFunctScale(mm);
        MPI_Bcast(&curv->WindowData->WindowFuncScale[0][0], curv->WindowData->NgridsWinScale[mm], MPI_FLOAT, (ncpus-1), MPI_COMM_WORLD);

        sprintf(FileName, "./CCoeff/%s_cc_ind-k%d", InputFile, mm);  sprintf(FileName1, "./CCoeff/%s_cc_val-k%d", InputFile, mm);  sprintf(FileName2, "./CCoeff/%s_cc_info-k%d", InputFile, mm);  
        curv->parReadCcoeff(FileName, FileName1, FileName2, &ivectv, &ivect1, icc_angles, Nangles, MPI_COMM_WORLD);

        curv->InvTransformScaleQuadPar(&ivectv, &ivect1,  icc_angles, imag_rec, curv->WindowData->WindowFuncScale, PPbeg, PPend, mm, angl1, dangl1);        
 
       free(icc_angles); 
        DMath::del(curv->WindowData->WindowFuncScale);

   } // for(mm=1; ....

             //   complicated layout to reduce buffer sizes ;
       curv->parReduceRedMem(imag_rec, dNx, dNy, dNz, MPI_COMM_WORLD);

        if( my_rank == (ncpus-1))    curv->WriteDataBlock(imag_rec, PPbeg, PPend, 1); 
  
       DMath::del(imag_rec);  
}

MPI_Barrier(MPI_COMM_WORLD);
cout << "INVERSE Smearing: ... done.\n"; cout.flush();
MPI_Barrier(MPI_COMM_WORLD);


// ==============  PAR - Global InvFFT ===================
if(1>0) {

   MPI_Group  all_cpus, fft_cpus; 
   MPI_Comm FFT_Comm; 
   int fftncpus, tmp, color, n_rank;
   char mvFile2[1024]; sprintf(mvFile2, "./CCoeff/%s_max_v",InputFile);

   tmp=(int)(log((double)ncpus)/log(2.0));
   fftncpus=(int)pow(2,tmp); 
   if(my_rank<fftncpus)  color=1; else color=0;
   MPI_Comm_split( MPI_COMM_WORLD, color, my_rank, &FFT_Comm);
   MPI_Comm_rank(MPI_COMM_WORLD, &n_rank);

   if(my_rank<fftncpus)   {
      sprintf(FileName, "%s_rec", InputFile);  
      curv->parIFFT(FileName, mvFile2, FFT_Comm);  }
}


MPI_Barrier(MPI_COMM_WORLD);
cout << "INVERSE FFT: ... done.\n"; cout.flush();
MPI_Barrier(MPI_COMM_WORLD);

cout << "=====================INVERSE TRANSFORM DONE=============\n"; cout.flush();
MPI_Barrier(MPI_COMM_WORLD);

 cout << "execution time for cpu: " << my_rank << ": "<< (MPI_Wtime()-t1)  <<"\n";  cout.flush();    

  MPI_Finalize();
  return 0;
}


